import ply.lex as lex

class StringerLex:
    states = (
        ('macroGrp',    'exclusive'),     
        ('textGrp',     'exclusive'),   # macros for replacement
        ('dict',        'exclusive'),
        ('pair',        'exclusive'),
        ('array',       'exclusive'),
        ('string',      'exclusive'),
        ('code',        'exclusive'),
        ('macroCall',   'exclusive')
    )

    tokens = (
        'MacroGrp',
        'TextGrp',
        'Variable',         # Used for macro name and parameters. Ignores whitespaces
        'String',           # Substrings. Uses whitespaces
        'Code',             # Python code
        'Equal',
        'Comma',
        'Number',
        'Comment',
        
        'LBrace',
        'RBrace',
        'LSquare',
        'RSquare',
        'LParen',
        'RParen',
        'LAngle',
        'RAngle',
        'Quote',
        'CodeQuote',
        
        # Built in macros
        'MacroDot',
        'MacroPercent',
        'MacroLevel'
    )

    def t_macroGrp_textGrp_dict_array_INITIAL_newline(self, t):
        r'\n+'
        t.lexer.lineno += len(t.value)
        
    def t_INITIAL_macroGrp_textGrp_MacroGrp(self, t):
        r'macros:'
        t.lexer.begin('macroGrp')
        return t
        
    def t_INITIAL_macroGrp_textGrp_TextGrp(self, t):
        r'texts:'
        t.lexer.begin('textGrp')
        return t
        
    def t_macroGrp_textGrp_dict_CodeQuote(self, t):
        r'c"'
        t.lexer.push_state('code')
        return t
    
    def t_macroGrp_textGrp_dict_pair_array_macroCall_INITIAL_Comment(self, t):
        r'\#[^\n]*'
        pass
        
    def t_macroGrp_textGrp_dict_macroCall_Variable(self, t):
        r'[a-zA-Z/\.\%\+]+(_[a-zA-Z/\.\%\+]+)*'
        return t
        
    def t_string_String(self, t):
        r'([^\\<>"]|\\\\|\\<|\\>|\\")+'
        t.value = t.value.replace('\\\\', '\\')
        t.value = t.value.replace('\\<', '<')
        t.value = t.value.replace('\\>', '>')
        t.value = t.value.replace('\\"', '"')
        return t
        
    def t_macroGrp_textGrp_dict_Equal(self, t):
        r'='
        return t
        
    def t_dict_pair_array_macroCall_Comma(self, t):
        r','
        return t
        
    def t_macroGrp_dict_pair_array_macroCall_Number(self, t):
        r'(-)?[0-9]+(\.[0-9]+)?'
        if '.' in t.value:
            t.value = float(t.value)
        else:
            t.value = int(t.value)
        return t
        
    def t_macroGrp_dict_LBrace(self, t):
        r'\{'
        t.lexer.push_state('dict')
        return t
        
    def t_dict_RBrace(self, t):
        r'\}'
        t.lexer.pop_state()
        return t
        
    def t_macroGrp_dict_LSquare(self, t):
        r'\['
        t.lexer.push_state('array')
        return t
        
    def t_array_RSquare(self, t):
        r'\]'
        t.lexer.pop_state()
        return t
        
    def t_macroGrp_dict_LParen(self, t):
        r'\('
        t.lexer.push_state('pair')
        return t
        
    def t_pair_RParen(self, t):
        r'\)'
        t.lexer.pop_state()
        return t
        
    def t_string_macroCall_LAngle(self, t):
        r'<'
        t.lexer.push_state('macroCall')
        return t
        
    def t_macroCall_RAngle(self, t):
        r'>'
        t.lexer.pop_state()
        return t
        
    def t_macroGrp_textGrp_dict_array_macroCall_Quote(self, t):
        r'"'
        t.lexer.push_state('string')
        return t
        
    def t_string_code_Quote(self, t):
        r'"'
        t.lexer.pop_state()
        return t
        
        
    def t_macroCall_MacroDot(self, t):
        r'\.'
        return t
        
    def t_macroCall_MacroPercent(self, t):
        r'%'
        return t
        
    def t_macroCall_MacroLevel(self, t):
        r'_level'
        return t
        
    def t_code_Code(self, t):
        r'[^"]+'
        return t
    
    t_macroGrp_textGrp_dict_pair_array_macroCall_INITIAL_ignore  = ' \t\r'
    t_ANY_ignore = '\r'
    
    # Error handling rule
    def t_ANY_error(self, t):
        print "Illegal character '%s'" % t.value[0]
        t.lexer.skip(1)
        
    # Build the lexer
    def build(self, **kwargs):
        self.lexer = lex.lex(module=self, **kwargs)

    # Test it output
    def test(self, data):
        self.lexer.input(data)
        while True:
             tok = self.lexer.token()
             if not tok: break
             print tok

